/*  $Id: CObj.hh,v 1.10 1999/07/16 20:18:49 shaggy Exp $
 *      relative path: ./lib/cparse
 */

#ifndef __CObj_hh__
#define __CObj_hh__

#include <vStrTab>
#include <vSymTab>

// Note:
//	The entire object system found here is meant to be temporary
//	It was designed to provide some of the functionality of
//	the DC++ system of dynamic objects, and it was designed
//	prior to the completion of the dynamic object model used
//	by DC++. 
//
//	At some point, all of this is supposed to go away and
//	be replaced by a bunch of .dh files describing all the
//	core types found in C++.
//
//	When this happens, it means that DECO will depend upon
//	itself, but this is reasonable, since the previously
//	converted headers (.dh.h files) can be used to build
//	version of DECO that can compile the new headers.  In this
//	manner, new versions of DECO can be bootstrapped.



// placement new (should be in new.h)

inline void* operator new(size_t, void *p)  { return p; }

typedef size_t bitsize_t;


typedef class Object obj_t;
typedef class Object *type_t;


#define GEN_DCAST(t) inline t* dcast_##t(Object *o) \
	{ return (o && o->istype(Type_##t)) ? reinterpret_cast<t*>(o->ptr()) : 0; }
#define GEN_COPYCTOR(o, s) \
	o(const s &data) : Object(s::type(), Inline) { new(&_val)s(data); } 
#define GEN_REFCTOR(o, s) \
	o(s *data) : Object(s::type(), Auto) { _ref.ptr = data; }
#define GEN_COPYREFCTOR(o, s)  GEN_COPYCTOR(o, s) GEN_REFCTOR(o, s)


class Object {
public:
	enum StorageType {
		Inline = 0,	// data is stored inline in object descriptor
		Auto = 1,	// ptr to automatic storage 
		Alloc = 2,	// ptr to allocated storage (must free)
		_Reserved = 3	// reserved for future use
	};
protected:
	// the lower 2 bits of the type are used for the StorageType
	// thus all objects used for types must have 32-bit alignment
	vaddr_t _type;
	union {
		uint64	_val;
		double	_fval;
		struct {
			void *ptr;
			uint32 count;
		} _ref;
	};
	obj_t *_refcnt;	// note: some inline objects are 
			// currently flowing over into refcnt, since no
			// bounds checking is done
	Object(type_t type, StorageType st) 
		{ _type = (vaddr_t)type | st; } 
	Object(obj_t &type, StorageType st) 
		{ _type = (vaddr_t)&type | st; } 

public:
	Object()
		{ _type = 0; _val = 0; }

	// Perhaps all of these functions for operating on the object
	// descriptor (rather than the contained object) should be 
	// prefixed in some way to distinguish them?

	type_t type() const
		{ return (type_t)(_type & ~3); }
	StorageType storagetype() const
		{ return (StorageType)(_type & 3); }
	bool isnull() const
		{ return !type(); }
	bool istype(obj_t &t) const
		{ return type() == &t; }
	bool istype(type_t t) const
		{ return type() == t; }
	// bool isof() const;	// isa? is_a?  ISA?
	bool stinline() const { return storagetype() == Inline; } 
	bool stauto() const { return storagetype() == Auto; } 
	bool stalloc() const { return storagetype() == Alloc; } 
	void * ptr() { return storagetype() == Inline
			? (void*)(&_val) : _ref.ptr; } 
	const void * ptr() const { return storagetype() == Inline
			? (const void*)(&_val) : (const void*)_ref.ptr; } 
	uint64 data() const { return storagetype() == Inline
				? _val : *(uint64*)_ref.ptr; } 
	size_t _sizeof() const;
	obj_t operator () (obj_t obj);
	obj_t operator () (obj_t *obj);
	operator bool () { return !isnull(); }
};


#define OBJ_NULL	Object()





// this should be an object describing the type:
extern obj_t Type_Integral;


class Integral {
private:
	uint16	_bits;		// total number of bits
	uint8	_alignment;	// bit alignment, as a power of two
	uint8	_sign:1;	// true if signed
	uint8	_bigend:1;	// true if big endian
	uint8	_reserved:5;	// reserved
public:
	Integral(int bits, int alignment, bool sign, bool bigend);
	bool sign() const { return _sign; }
	size_t bitsize() const { return _bits; }
	size_t size() const { return ((_bits-1)/8)+1; }
	uint alignment() const { return _alignment; }
	bool bigendian() const { return _bigend; }
	bool lilendian() const { return !_bigend; }
	static type_t type() { return &Type_Integral; }
};


inline Integral::Integral(int bits, int alignment,
		bool sign, bool bigend) {
	_bits = bits;
	_alignment = alignment;	
	_sign = sign;
	_bigend = bigend;
	_reserved = 0;
}




////////////////////////////////

class obj_Integral : public Object {
public:
	obj_Integral(int bits, int alignment, bool sign,
		bool bigend) : Object(Integral::type(), Inline)
		{ new(&_val)Integral(bits, alignment, sign, bigend); }
	operator Integral() { return *(Integral*)ptr(); }
	GEN_COPYREFCTOR(obj_Integral, Integral)
};

GEN_DCAST(Integral)

////////////////////////////////////////////////////////

extern obj_t Type_PtrTo;

class PtrTo {
	const type_t _type;
	bool _qconst:1;
	bool _qvolatile:1;	
public:
	// obj_t* operator*() { return ptr; }
	PtrTo(type_t type, bool c = false, bool v = false) : 
			_type(type), _qconst(c), _qvolatile(v) {}
	type_t rtype() const { return _type; }
	bool qconst() { return _qconst; }
	bool qvolatile() { return _qvolatile; }
	static type_t type() { return &Type_PtrTo; }
};


class obj_PtrTo : public Object {
public:
	obj_PtrTo(type_t type, bool c = false, bool v = false)
			: Object(PtrTo::type(), Inline)
		{ new(&_val)PtrTo(type, c, v); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_PtrTo, PtrTo)
};


GEN_DCAST(PtrTo)

////////////////////////////////////////////////////////

extern obj_t Type_RefTo;

class RefTo {
	const type_t _type;
	bool _qconst:1;
	bool _qvolatile:1;	
public:
	// obj_t* operator*() { return ptr; }
	RefTo(type_t type, bool c = false, bool v = false) :
			_type(type), _qconst(c), _qvolatile(v) {}
	type_t rtype() const { return _type; }
	bool qconst() { return _qconst; }
	bool qvolatile() { return _qvolatile; }
	static type_t type() { return &Type_RefTo; }
};


class obj_RefTo : public Object {
public:
	obj_RefTo(type_t type, bool c = false, bool v = false)
			: Object(RefTo::type(), Inline)
		{ new(&_val)RefTo(type, c, v); }
	GEN_COPYREFCTOR(obj_RefTo, RefTo)
};

GEN_DCAST(RefTo)

////////////////////////////////////////////////////////

extern obj_t Type_Qualifier;

class Qualifier {
	const type_t _type;
	bool _qconst:1;
	bool _qvolatile:1;	
public:
	Qualifier(type_t type, bool c = false, bool v = false) : 
			_type(type), _qconst(c), _qvolatile(v) {}
	type_t rtype() const { return _type; }
	bool qconst() { return _qconst; }
	bool qvolatile() { return _qvolatile; }
	static type_t type() { return &Type_Qualifier; }
};


class obj_Qualifier : public Object {
public:
	obj_Qualifier(type_t type, bool c = false, bool v = false)
			: Object(Qualifier::type(), Inline)
		{ new(&_val)Qualifier(type, c, v); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Qualifier, Qualifier)
};

GEN_DCAST(Qualifier)

////////////////////////////////////////////////////////

extern obj_t Type_Array;


class Array {
	const type_t _type;
	const size_t _size;
	void *_dsize;
public:
	Array(type_t type, size_t size, void *dsize = 0)
		: _type(type), _size(size), _dsize(dsize) {}
	size_t size() const
		{ return _size; }
	void* dsize() const
		{ return _dsize; }
	size_t lower() const { return 0; }
	size_t upper() const { return _size-1; }
	type_t rtype() const { return _type; }
	static type_t type() { return &Type_Array; }
};



class obj_Array : public Object {
public:
	obj_Array(type_t type, size_t size, void *dsize = 0)
		: Object(Array::type(), Inline)
		{ new(&_val)Array(type, size, dsize); }
	GEN_COPYREFCTOR(obj_Array, Array)
};

GEN_DCAST(Array)

////////////////////////////////////////////////////////


// Params defines the input signature for a function
// and Func defines the complete I/O signature (Param and return type)

// Note: Limited to 255 specified parameters for any function
// (and if any function needs more than about 8, something is probably 
// terribly wrong with the design and organization of the
// function.  Most should never need more than 2 or 3 max...
// Remember that the human mind usually only keeps 7 plus or minus 2
// items in short term memory, so some sort of chunking should be
// used for keeping track of any more than that.
// Okay, I'll get off my soapbox now...)


extern obj_t Type_Params;

class Params {
private:
	const type_t	*_ptype;
	const uint8	_nparam;
	const bool 	_variable:1;
public:
	Params(int nparam, const type_t *ptype, 
		bool variable = false)
		      : _ptype(ptype), _nparam(nparam),
			_variable(variable) {}
	bool variable() const { return _variable; }
	uint nparam() const { return _nparam; }
	type_t ptype(uint index) const { return _ptype[index]; }
	static type_t type() { return &Type_Params; }
};

class obj_Params : public Object {
public:
	obj_Params(int nparam, const type_t *ptype, 
		bool variable = false)
		: Object(Params::type(), Inline)
		{ new(&_val)Params(nparam, ptype, variable); }
	GEN_COPYREFCTOR(obj_Params, Params)
};

GEN_DCAST(Params)

extern obj_t Type_Func;

class Func {
private:
	const type_t	_rtype;
	const type_t	_psig;
	vSymTab		*_scope;
public:
	Func(type_t rtype, type_t param, vSymTab *scope)
		      : _rtype(rtype), _psig(param), _scope(scope) {}
	bool variable() const { return ((Params*)_psig->ptr())->variable(); }
	uint nparam() const { return ((Params*)_psig->ptr())->nparam(); }
	type_t ptype(uint n) const { return ((Params*)_psig->ptr())->ptype(n); }
	type_t rtype() const { return _rtype; }
	type_t psig() const { return _psig; }
	vSymTab *scope() const { return _scope; }
	static type_t type() { return &Type_Func; }
	void change_rtype(type_t type) { *(type_t*)&_rtype = type; } /* XXX - HACK */
	void change_scope(vSymTab *scope) { _scope = scope; } /* XXX - HACK */
};


class obj_Func : public Object {
public:
	obj_Func(type_t rtype, type_t param, vSymTab *scope)
		: Object(Func::type(), Inline)
		{ new(&_val)Func(rtype, param, scope); }
	GEN_COPYREFCTOR(obj_Func, Func)
};

GEN_DCAST(Func)

////////////////////////////////////////////////////////

extern obj_t Type_Method;

class Method {
	const type_t _otype;
	const type_t _ftype;
public:
	Method(type_t otype, type_t ftype) :
			_otype(otype), _ftype(ftype) {}
	type_t ftype() const { return _ftype; }
	type_t otype() const { return _otype; }
	static type_t type() { return &Type_Method; }
};


class obj_Method : public Object {
public:
	obj_Method(type_t otype, type_t ftype)
			: Object(Method::type(), Inline)
		{ new(&_val)Method(otype, ftype); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Method, Method)
};

GEN_DCAST(Method)

////////////////////////////////////////////////////////


extern obj_t Type_Aggr;

struct ObjAttr {
	vStrTab::Id	name;		// or 0 for private attributes
	type_t		type;
	bitsize_t	offset;		
	bitsize_t	size;		// XXX - or make special types?
};

struct AggrAttr {
	vStrTab::Id	name;
	obj_t		val;
};


class Aggr {
	vStrTab::Id	_name;
	// bitsize_t	size;
	// uint16		nobj_attr;
	// uint16		nclass_attr;
	// ObjAttr		*obj_attr;
	// AggrAttr	*class_attr;
	vSymTab		*_scope;
public:
	Aggr(vStrTab::Id name, vSymTab *scope) : _name(name), _scope(scope) {}
	vStrTab::Id name() const { return _name; }
	vSymTab *scope() const { return _scope; }
	void scope(vSymTab *scope) { _scope = scope; }
	static type_t type() { return &Type_Aggr; }
};

class obj_Aggr : public Object {
public:
	obj_Aggr(vStrTab::Id name, vSymTab *scope)
			: Object(Aggr::type(), Auto)
		{ _ref.ptr = new Aggr(name, scope); }
	size_t _sizeof() const;
	GEN_COPYREFCTOR(obj_Aggr, Aggr)
};

GEN_DCAST(Aggr)

////////////////////////////////////////////////////////

extern obj_Integral Type_char;
extern obj_Integral Type_schar;
extern obj_Integral Type_int;
extern obj_Integral Type_short;
extern obj_Integral Type_long;
extern obj_Integral Type_long2;
extern obj_Integral Type_uchar;
extern obj_Integral Type_uint;
extern obj_Integral Type_ushort;
extern obj_Integral Type_ulong;
extern obj_Integral Type_ulong2;

extern obj_t	Type_void;
extern obj_t	Type_bool;

extern obj_t	Type_ptr;
extern obj_t	Type_ref;

extern obj_t	Type_aggr;

extern obj_t	Type_List;
extern obj_t	Type_TokenList;

extern obj_t	Type_Str;

extern obj_t	Type_wchar;
extern obj_t	Type_float;
extern obj_t	Type_double;

extern obj_t	Type_null;	// could be (int)0 or (AnyType*)0
				// personally, I think that allowing 0
				// to be equivalent to null was a very
				// unwise move in terms of type-safety,
				// and creating a new keyword of
				// 'null' or 'nil' would have been
				// preferable to allowing this implicit
				// conversion. 


#endif
