/*  $Id: CParse.cc,v 1.39 1999/07/16 20:18:49 shaggy Exp $
 *      C/C++ Parser (yacc grammar)
 *      
 *      relative path: lib/cparse
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define IMPLEMENTATION_CParse

#include <libc_port.h>
#include <stdx1.h>
#include <defines>
#include <ldefines>

#include "CParse.hh"

#define YYSTYPE vCAction::Item


// #define none vCAction::none


void
CParse::_internal_error(vErrorReport *er, csz *file, int line)
{ er->internal_error(file, line, "CParse"); }


CParse::CParse(CLex &clex, vCAction &caction) 
		: _clex(&clex), _caction(&caction)
{
	_prev_yychar = 0;
	_havenext = false;
	_current_scope = CLex::NONE;
	_stmtlist = CLex::NONE;
	_debug = false;
	_yydebug = false;
	_rawid = false;
	_gsavecode = _savecode = false;
	_deftok = 0;
	_saveline = 0;
}

void
CParse::inject(TokenList *tok)
{
	if(_toksrc)
		tok->append_delete(_toksrc);
	_toksrc = tok;
}

void
CParse::newdecl(Item *list, Item_ decl)
{
	if(list->token != T(NONE) && list->scope == _current_scope.scope) {
		Item stmt = decl;
		stmt.token = T(DECL);
		do3(*list, BuildStmtList, *list, stmt, _current_scope);
	}
} 


// XXX - some of this functionality probably needs to be offloaded
//	to CAction, to allow greater flexibility
//	since it's not truly part of the parser

int
CParse::_lex(YYSTYPE *yylval)
{
	while(true) {
		if(_havenext) {
			*yylval = _next_yylval;
			_havenext = false;
			return _next_yychar;
		} else if(_havenextnext) {
			*yylval = _nextnext_yylval;
			_havenextnext = false;
			return _nextnext_yychar;
		} 

		if(_toksrc) {
			if(!_saveline)
				_saveline = _clex->line_num();
			if(_toksrc->next(yylval)) {
				_clex->set_line_num(yylval->line);
				return yylval->token;
			} else {
				delete _toksrc;
				_toksrc = 0;
				_clex->set_line_num(_saveline);
				_saveline = 0;
			}
		}

		yylval->val = 0;
		vCAction::Token token = (vCAction::Token)_clex->next_token();
		yylval->token = token;
		switch(token) {
		case T(NONE):
			_caction->get_input(false); continue;
		case T(INCOMPLETE):
			_caction->get_input(true); continue;
		case T(EOF_COMPLETE):
			return 0;
		case T(EOF_INCOMPLETE):
			if(_clex->in_string())
				_caction->error("unterminated string");
			else if(_clex->in_char())
				_caction->error("unterminated character constant");
			else if(_clex->in_bquote())
				_caction->error("unterminated backquote");
			else
				_caction->error("incomplete final token");
			return 0;
		case T(INVALID):
			_caction->error("invalid character"); continue;
		case T(NUM):
			yylval->Num = _clex->last_num();
			yylval->c.sign = _clex->last_signed();
			yylval->c.nlong = _clex->last_long();
			break;
		case T(BOOL):
			yylval->Bool = _clex->last_bool(); break;
		case T(KW_TRUE):
			token = yylval->token = T(BOOL);
			yylval->Bool = true;
			break;
		case T(KW_FALSE):
			token = yylval->token = T(BOOL);
			yylval->Bool = false; 
			break;
		case T(STR):
		case T(BACKSTR):
		case T(ANGSTR):
		case T(ID):
			yylval->id = _caction->id(
				token, _clex->last_str()); break;
		case T(CHAR):
			yylval->Char = _clex->last_char(); break;
		case T(FLOAT):
			yylval->Float = _clex->last_float(); break;
		case T(EOL):
			_caction->eol(); continue;
		case T(KW__FUNC):
		case T(KW__PFUNC):
			do2(*yylval, SpecialStr, *yylval, _current_scope);
			token = yylval->token;
			break;
		default: break;
		}
		yylval->line = _clex->line_num();
		// printf("sending yychar %d\n", (int)token);
		return (int)token;
	}
}


#define set_next_yy(c, l) \
	{ _next_yychar = (c); _next_yylval = (l); _havenext = true; }
#define set_nextnext_yy(c, l) \
	{ _nextnext_yychar = (c); _nextnext_yylval = (l); _havenextnext = true; }


// peekahead_lex() allows peaking ahead in the token stream
// for complex sequences.
// It handles scoping with the '::' operator,
// as well as looking up identifiers to find their types and
// aid the parser

int
CParse::_peekahead_lex(YYSTYPE *yylval)
{
	Item scope;
	bool isdtor;
	int yychar = _lex(yylval);

	if(_rawid)
		goto ReturnToken;

	// printf("got %d,%d\n", yychar, T(KW_TEMPLATE));

	// We can't support templates yet...
	if(yychar == T(KW_TEMPLATE))
		_caction->abort("templates are not yet supported");

	// We can't support exception handling yet
	if(yychar == T(KW_THROW) || yychar == T(KW_CATCH)
			|| yychar == T(KW_TRY))
		_caction->abort("exception handling not yet supported");

	if(yychar == T(OP_NEW) || yychar == T(OP_DELETE)) {
		Item nyylval;
		int nyychar = _lex(&nyylval);
		if(nyychar == T(LBRACK)) {
			Item nnyylval;
			int nnyychar = _lex(&nnyylval);
			if(nnyychar == T(RBRACK)) {
				yylval->token = yylval->token == T(OP_NEW)
					? T(OP_NEWA) : T(OP_DELETEA);
				yychar = yylval->token;
				goto ReturnToken;
			}
			set_nextnext_yy(nnyychar, nnyylval);
		} 
		set_next_yy(nyychar, nyylval);
	}

	if((isdtor = (yychar == T(OP_NOT)))) {
		yychar = _lex(yylval);
		if(yychar != T(ID)) {
			set_next_yy(yychar, *yylval);
			yychar = T(OP_NOT); *yylval = T(OP_NOT);
		}
	}

	scope.token = T(SCOPE);
	scope.scope = 0;
	if(yychar == T(OP_SCOPE)) {
		yychar = _lex(yylval);
		do3(scope, FindSym, T(SCOPE), T(NONE), none);
		if(yychar != T(ID)) {
			set_next_yy(yychar, *yylval);
			*yylval = scope;
			yychar = T(SCOPE);
		}
	}

	/* scope field of ID should be null unless a scope
 	 * is explicitly stated, in which case it should point
	 * to that symtab 
			
	 */
	if(yychar == T(ID)) {
		while((_next_yychar = _lex(&_next_yylval)) == T(OP_SCOPE)) {
			if(isdtor) {
				set_nextnext_yy(_next_yychar, _next_yylval);
				set_next_yy(yychar, *yylval);
				yychar = T(OP_NOT); *yylval = T(OP_NOT);
				goto ReturnToken;
			}
			yylval->scope = scope.scope;
			do3(scope, FindSym, T(SCOPE), *yylval, _current_scope);
			yychar = _lex(yylval);
			if(scope.token == T(NONE)) {
				_caction->error("undefined scope");
				scope.scope = 0;
			}
			if(yychar != T(ID)) {
				set_next_yy(yychar, *yylval);
				*yylval = scope;
				yychar = T(SCOPE);
				goto ReturnToken;
			}
		}

		_havenext = true;

		Item def;
		Item type;

		if(_prev_yychar == T(KW_CLASS) || _prev_yychar == T(KW_STRUCT)
		 || _prev_yychar == T(KW_UNION) || _prev_yychar == T(KW_ENUM))
			type = (CLex::Token)(_prev_yychar);
		else
			type = T(NONE);

		yylval->scope = scope.scope;	
		if(yychar != T(ID))
			goto ReturnToken;
		do3(def, FindSym, type, *yylval, _current_scope);

		/* XXX - hack */
		if(def.token == T(CTOR_ID) && _next_yychar != T(LPAREN))
			def.token = T(AGGR_ID);
	
		switch(def.token) {
		case T(TYPE_ID):
		case T(ENUM_ID):
		case T(NSPACE_ID):
		case T(ASPACE_ID):
		case T(TPLATE_ID):
		case T(AGGR_ID):
		case T(VAL_ID):
			yychar = def.token;
			*yylval = def;
			break;
		case T(CTOR_ID):
			yychar = isdtor ? T(DTOR_ID) : def.token;
			*yylval = def;
			break;
		default:
			break;
		}

		if(isdtor && yychar != T(DTOR_ID)) {
			set_nextnext_yy(_next_yychar, _next_yylval);
			set_next_yy(yychar, *yylval);
			yychar = T(OP_NOT); *yylval = T(OP_NOT);
		}
	}

ReturnToken:		 
	if(_debug)
		_caction->eprintf("next token = %s\n",
				_caction->cstr((vCAction::Token)yychar));
	_prev_yychar = yychar;

	if(_savecode) {
		if(yychar == T(LBRACE))
			yychar = T(LBRACE_SAVE);
		else if(yychar == T(COLON))
			yychar = T(COLON_SAVE);
	}
	return yychar;
}

void
CParse::_warn_unused_type(Item_ t)
{
	_caction->warning("unused type declaration: %s\n", _caction->cstr(t.token));
}

int
CParse::parse()
{
	if(_do_parse()) {
		_caction->abort("cannot recover from earlier errors");
		return 0;
	} else {
		_caction->accept();
		return 1;
	}
}

