/*  $Id: CRep.hh,v 1.26 1999/07/16 20:18:49 shaggy Exp $
 *      C++ internals (knowledge representation and manipulation)
 *      relative path: lib/cparse
 *  
 *  Copyright (c) 1996-1999 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __CRep_hh__
#define __CRep_hh__

#include <CObj>
#include <CLex>
#include <SymTab>
#include <StrTab>
#include <vCAction>

#define T(x)	vCAction::x

struct Decl;
class TokenList;

class Scope {	// more like a ScopePtr or ScopeRef actually...
	vSymTab *_sym;
public:
	typedef vSymTab::Id Id;
	Scope() { _sym = 0; }
	Scope(vSymTab *s) { _sym = s; }
	Scope(const Scope &s) { _sym = s._sym; }
	const Scope& operator= (vSymTab *s) const
		{ const_cast<vSymTab*&>(_sym) = s; return *this; }
	const Scope& operator= (Scope s) const
		{ const_cast<vSymTab*&>(_sym) = s._sym; return *this; }
	operator vSymTab* () const { return _sym; }
	bool operator== (Scope s) const { return _sym == s._sym; }
	bool operator!= (Scope s) const { return _sym != s._sym; }
	bool operator== (vSymTab *s) const { return _sym == s; }
	bool operator!= (vSymTab *s) const { return _sym != s; }
	Decl* get(Id id) const
		{ return reinterpret_cast<Decl*>(_sym->get(id)); }
	bool set(Id id, Decl *d) { return _sym->set(id, d); }
	bool add(Id id, Decl *d) { return _sym->add(id, d); }
	bool del(Id id) { return _sym->del(id); }
	Id index(uint n) const { return _sym->index(n); }
	Scope parent() const { return Scope(_sym->parent()); }
	bool parent(Scope s) { return _sym->parent(s); }
	// Scope transplant() { return Scope(_sym->transplant()); }
	Decl* decl() const
		{ return reinterpret_cast<Decl*>(_sym->data()); }
	bool decl(Decl *d)
		{ return _sym->data(d); }
	Id pchid() const;
	vStrTab* strtab() const { return _sym->strtab(); }
	operator bool () const { return _sym != 0; }

	// untagged lookup:
	inline Decl* uget(vStrTab::Id id) const;
	inline Decl* ufind(vStrTab::Id id) const;

	// tagged lookup:
	inline Decl* tget(vStrTab::Id id) const;
	inline Decl* tget(vStrTab::Id id, bool t) const;
	inline Decl* tfind(vStrTab::Id id) const;

	// selectable tagged lookup:
	inline Decl* tfind(vStrTab::Id id, bool t) const;
	inline Decl* tfind(vStrTab::Id id, bool t, bool s) const;
};

// templates could be used to do this generically, but then
// deco wouldn't be able to parse itself, so until deco has
// template support, none of the code shall use templates
class ScopeList {
	struct Node {
		vSymTab *sym;
		Node *next;
	};
	Node *_first;
	Node *_last;
public:
	ScopeList() { _first = _last = 0; }
	ScopeList(Node *first, Node *last) { _first = first; _last = last; }
	Scope head() const { return Scope(_first ? _first->sym : 0); }
	ScopeList tail() const { return _first ? ScopeList(_first->next, _last)
				 : ScopeList(0, 0); }
	void prepend(Scope s) { Node *n = new Node; n->sym = s;
				n->next = _first;
				if(!_last) _last = n;
				_first = n; }
	void append(Scope s) { Node *n = new Node; n->sym = s;
				n->next = 0;
				if(_last) _last->next = n; 
				else _first = n;
				_last = n; }
	ScopeList& operator += (Scope s) { append(s); return *this; }
};

// currently unused:
class DeclList {
	struct Node {
		Decl *decl;
		Node *next;
	};
	Node *_first;
	Node *_last;
public:
	DeclList() { _first = _last = 0; }
	DeclList(Node *first, Node *last) { _first = first; _last = last; }
	Decl* head() const { return _first ? _first->decl : 0; }
	DeclList tail() const { return _first ? DeclList(_first->next, _last)
				 : DeclList(0, 0); }
	void prepend(Decl *d) { Node *n = new Node; n->decl = d;
				n->next = _first;
				if(!_last) _last = n;
				_first = n; }
	void append(Decl *d) { Node *n = new Node; n->decl = d;
				n->next = 0;
				if(_last) _last->next = n; 
				else _first = n;
				_last = n; }
	DeclList& operator += (Decl *d) { append(d); return *this; }
};


struct Litstruct {
	type_t	type;
	union {
		uint64 val;
		void *ptr;
	};
};

class Lit : public Litstruct {
public:
	Lit(type_t t, uint64 v) { type = t; val = v; }
	Lit(type_t t, int v) { type = t; val = v; }
	Lit(type_t t, void *p) { type = t; ptr = p; }
	bool operator==(const Lit &n) const
		{ return type == n.type && val == n.val; }
	bool operator!=(const Lit &n) const { return !(*this == n); }
};

struct ExprNode {
	vCAction::Token op;
	union {
		struct {
			ExprNode *left;
			union {
				ExprNode *mid;
				vSymTab *sc_mid;
				vCAction::Token tok_mid;
			};
			ExprNode *right;
		} e;
		Litstruct lit;
		Decl *decl;
		TokenList *tlist;
		// vSymTab *scope;
		vStrTab::Id id;
	};
	type_t rtype;
	ExprNode *next;
};

class Expr {
	ExprNode *p;
	friend class ExprList;
	friend class Stmt;
public:
	Expr() { p = 0; }
	Expr(ExprNode *np) { p = np; }
	Expr(const Lit &);
	Expr(Decl *);
	Expr(type_t);
	Expr(TokenList *);
	// Expr(const Scope scope);
	Expr(vStrTab::Id);
	Expr(vCAction::Token tok, Expr e);
	Expr(vCAction::Token tok, Expr e1, Expr e2);
	Expr(vCAction::Token tok, Expr e1, Expr e2, Expr e3);
	Expr& operator+= (Expr e1)
		{ p->next = e1.p; return *this; }
	Expr operator+ (Expr e1) const
		{ Expr e(*this); e.p->next = e1.p; return e; }
	bool operator== (Expr) const;
	bool operator!= (Expr e) const { return !(*this == e); }
	bool is(Expr e) const { return p == e.p; }
	Expr& rtype(type_t t) { p->rtype = t; return *this; }
	vCAction::Token op() const { return p->op; }
	const Lit& lit() const { return (Lit&)p->lit; }
	Decl* decl() const { return p->decl; }
	TokenList* tlist() const { return p->tlist; }
	vStrTab::Id id() const { return p->id; }
	const Expr left() const { return p->e.left; }
	const Expr mid() const { return p->e.mid; }
	const Expr right() const { return p->e.right; }
	type_t rtype() const { return p->rtype; }
	~Expr() {  }
	operator ExprNode*() const { return p; }
	operator bool () const { return p != 0; }
	Expr next() const { return Expr(p->next); }
	Expr& operator++() { p = p->next; return *this; }
	Expr operator++(int) { Expr e = *this; p = p->next; return e; }
	Expr dup() const;

	// debugging:
	void dump(int indent = 0, csz* prefix = "") const;
};

static const Expr null_expr = Expr((ExprNode*)0);

class ExprList {
	ExprNode *p;
public:
	ExprList(vCAction::Token tok = T(EXPRLIST));
	ExprList(vCAction::Token tok, vCAction::Token fmt);
	ExprList(ExprNode *np) { p = np; }
	ExprList& append(Expr e);
	ExprList& append(ExprList e);
	ExprList& operator += (Expr e) { return append(e); }
	ExprList& operator += (ExprList e) { return append(e); }
	operator bool () const { return p != 0; }
	bool empty() const { return !p->e.left; }
	Expr first() const { return Expr(p->e.left); }
	Expr last() const { return Expr(p->e.right); }
	vCAction::Token fmt() const { return p->e.tok_mid; }
	int len() const;
	operator ExprNode*() const { return p; }
};

class Stmt {
	ExprNode *p;
	friend class StmtList;
public:
	Stmt(ExprNode *np) { p = np; }
	Stmt(const Lit &);
	Stmt(Decl *);
	Stmt(TokenList *);
	Stmt(vCAction::Token tok, Expr e);
	Stmt(vCAction::Token tok, Expr e1, Expr e2);
	Stmt(vCAction::Token tok, Expr e1, Expr e2, Expr e3);
	vCAction::Token op() const { return p->op; }
	const Lit& lit() const { return (Lit&)p->lit; }
	Decl* decl() const { return p->decl; }
	TokenList* tlist() const { return p->tlist; }
	const Expr left() const { return p->e.left; }
	const Expr mid() const { return p->e.mid; }
	const Expr right() const { return p->e.right; }
	~Stmt() {  }
	operator ExprNode*() const { return p; }
	operator bool () const { return p != 0; }
	Stmt next() const { return Stmt(p->next); }
	Stmt& operator++() { p = p->next; return *this; }
	Stmt operator++(int) { Stmt e = *this; p = p->next; return e; }
};

static const Stmt null_stmt = Stmt((ExprNode*)0);

class StmtList {
	ExprNode *p;
public:
	StmtList(vCAction::Token tok = T(STMTLIST), Scope scope = 0);
	StmtList(Scope scope, vCAction::Token tok = T(STMTLIST));
	StmtList(ExprNode *np) { p = np; }
	StmtList& append(Stmt e);
	StmtList& operator += (Stmt e) { return append(e); }
	operator bool () const { return p != 0; }
	Scope scope() const { return p->e.sc_mid; }
	bool empty() const { return !p->e.left; }
	Stmt first() const { return Stmt(p->e.left); }
	Stmt last() const { return Stmt(p->e.right); }
	operator ExprNode*() const { return p; }
};

class StmtIf : public Stmt { public: StmtIf(Expr c, Stmt s)
		: Stmt(T(KW_IF), c, Expr(s), null_expr) {} };
class StmtIfElse : public Stmt { public: StmtIfElse(Expr c, Stmt s, Stmt e)
		: Stmt(T(KW_IF), c, Expr(s), Expr(e)) {} };
class StmtWhile : public Stmt { public: StmtWhile(Expr c, Stmt s)
		: Stmt(T(KW_WHILE), c, Expr(s), null_expr) {} };
class StmtDoWhile : public Stmt { public: StmtDoWhile(Stmt s, Expr c)
		: Stmt(T(KW_DO), null_expr, Expr(s), c) {} };
class ExprFor : public Expr { public: ExprFor(Stmt i, Expr c, Expr n)
		: Expr(T(KW_FOR), Expr(i), c, n) {} };
class StmtFor : public Stmt { public: StmtFor(Stmt i, Expr c, Expr n, Stmt s)
		: Stmt(T(KW_FOR), ExprFor(i, c, n), Expr(s), null_expr) {} 
		StmtFor(Expr fe, Stmt s)
		: Stmt(T(KW_FOR), fe, Expr(s), null_expr) {} };
class StmtSwitch : public Stmt { public: StmtSwitch(Expr v, Stmt s)
		: Stmt(T(KW_SWITCH), v, Expr(s), null_expr) {} };
class StmtReturn : public Stmt { public: StmtReturn(Expr r)
		: Stmt(T(KW_RETURN), r, null_expr, null_expr) {} };
class StmtGoto : public Stmt { public: StmtGoto(Expr w)
		: Stmt(T(KW_GOTO), w, null_expr, null_expr) {} };
class StmtBreak : public Stmt { public: StmtBreak()
		: Stmt(T(KW_BREAK), null_expr, null_expr, null_expr) {} };
class StmtContinue : public Stmt { public: StmtContinue()
		: Stmt(T(KW_CONTINUE), null_expr, null_expr, null_expr) {} };
class StmtAsm : public Stmt { public: StmtAsm(Expr s)
		: Stmt(T(KW_ASM), s, null_expr, null_expr) {} };
class StmtLabel : public Stmt { public: StmtLabel(Expr n)
		: Stmt(T(COLON), n, null_expr, null_expr) {} 
		StmtLabel(Expr n, Stmt s)
		: Stmt(T(COLON), n, Expr(s), null_expr) {} };
class StmtCase : public Stmt { public: StmtCase(Expr n)
		: Stmt(T(KW_CASE), n, null_expr, null_expr) {} 
		StmtCase(Expr n, Stmt s)
		: Stmt(T(KW_CASE), n, Expr(s), null_expr) {} };
class StmtDefault : public Stmt { public: StmtDefault()
		: Stmt(T(KW_DEFAULT), null_expr, null_expr, null_expr) {} 
		StmtDefault(Stmt s)
		: Stmt(T(KW_DEFAULT), null_expr, Expr(s), null_expr) {} };
class ExprDecl : public Expr { public: ExprDecl(Decl *d)
		: Expr(T(DECL), Expr(d), null_expr, null_expr) {} };
class StmtDecl : public Stmt { public: StmtDecl(Decl *d)
		: Stmt(T(DECL), Expr(d), null_expr, null_expr) {} };
class ExprCall : public Expr { public: ExprCall(Expr c, ExprList e)
		: Expr(T(OP_FUNC), c, null_expr, Expr(e)) {} };
class StmtCall : public Stmt { public: StmtCall(Expr c, ExprList e)
		: Stmt(T(OP_FUNC), c, null_expr, Expr(e)) {} };
class ExprCtor : public Expr { public: ExprCtor(StmtList c, StmtList s)
		: Expr(T(CTORINIT), Expr(c), Expr(s), null_expr) {} };
class ExprCode : public Expr {
public:
	ExprCode(Scope /*sc*/, StmtList c, StmtList s)
		: Expr(T(CTORINIT), Expr(c), Expr(s), null_expr) {}
};


class DeclCommon {
public:
	enum Storage /* formerly known as StorageClass */ {
		sc_default = 0,
		sc_auto = 1,
		sc_static = 2,
		sc_extern = 3,
		sc_register = 4,
		sc_typedef = 5,
		sc_mutable = 6,
		sc_constant = 7,
		sc_param = 8,
		sc_friend = 9,
		sc_extern_c = 10,
		sc_extern_cxx = 11,
	};
	enum DataType {
		dt_default = 0,
		dt_int = 1,
		dt_char = 2,
		dt_wchar = 3,
		dt_bool = 4,
		dt_void = 5,
		dt_float = 6,
		dt_double = 7,
		dt_enum = 8,
		dt_struct = 9,
		dt_class = 10,
		dt_union = 11,
		dt_namespace = 12,
		dt_enumval = 13,	// constant from an enum
		dt_type = 14,		// type interface (virtual class)
		dt_dtype = 15,		// dynamic type interface
		dt_dclass = 16,		// dynamic class
		dt_impl = 17,		// implementation for a type
	};
	enum DataTypeFlags {
		DTF_SIGNOK = 0x01,
		DTF_SHORTOK = 0x02,
		DTF_LONGOK = 0x04,
		DTF_SCOPE = 0x08,	// defines a new scope
		DTF_AGGR = 0x10,
		DTF_TAG = 0x20,		// tagged types can co-exist in symtab
	};
	enum Access {
		ac_public = 0,
		ac_protected = 1,
		ac_private = 2,
		ac_unspecified = 3,
	};

protected:
	// note: the uchars below used to be bools, but there's this
	// problem with the struct getting aligned to the largest
	// contained alignment.  On Sparcs, bool is 32-bits (and
	// aligned as such), so even though we're only using one bit
	// for each bool, and 16-bits total here, this structure
	// gets rounded up to 32 bits, causing TDecl to be more than 64.
	DataType	_dtype:5;	ACCESSORS(DeclCommon, DataType, dtype)
	Access		_access:2;	ACCESSORS(DeclCommon, Access, access)
	char 		_rsvd:1;
	Storage		_storage:4;	ACCESSORS(DeclCommon, Storage, storage)
	uchar		_sinline:1;	ACCESSORS(DeclCommon, bool, sinline)
	uchar		_svirtual:1;	ACCESSORS(DeclCommon, bool, svirtual)
	uchar		_sexplicit:1;	ACCESSORS(DeclCommon, bool, sexplicit)
	uchar		_rsvd2:1;

	static csz* _scname[];
	static csz* _dtname[];
	static csz* _acname[];
	static uint8 _dtflags[];

public:
	static csz* scname(Storage s) { return _scname[s]; }
	static csz* dtname(DataType d) { return _dtname[d]; }
	static csz* acname(Access a) { return _acname[a]; }
	csz* scname() const { return scname(storage()); }
	csz* dtname() const { return dtname(dtype()); }
	csz* acname() const { return acname(access()); }

	bool tagged() const { return _dtflags[dtype()] & DTF_TAG; }
	bool dtaggr() const { return _dtflags[dtype()] & DTF_AGGR; }
	bool dtscope() const { return _dtflags[dtype()] & DTF_SCOPE; }
	bool dtlongok() const { return _dtflags[dtype()] & DTF_LONGOK; }
	bool dtshortok() const { return _dtflags[dtype()] & DTF_SHORTOK; }
	bool dtsignok() const { return _dtflags[dtype()] & DTF_SIGNOK; }
	Access defacc() const
		{ return dtype() == dt_class ? ac_private : ac_public; }
	bool external() const
		{ return storage() == sc_extern || storage() == sc_extern_c 
			|| storage() == sc_extern_cxx; }
};

class TDecl : public DeclCommon {
private:
	uint	_slong:3;	ACCESSORS(TDecl, uint, slong)
	bool	_sshort:1;	ACCESSORS(TDecl, bool, sshort)
	bool	_sunsigned:1;	ACCESSORS(TDecl, bool, sunsigned)
	bool	_ssigned:1;	ACCESSORS(TDecl, bool, ssigned)
	bool	_qconst:1;	ACCESSORS(TDecl, bool, qconst)
	bool	_qvolatile:1;	ACCESSORS(TDecl, bool, qvolatile)
	bool	_useful:1;	ACCESSORS(TDecl, bool, useful)
	bool	_rsvd:7;
	type_t	_ptype;		// printable type (based on definition)
				ACCESSORS(TDecl, type_t, ptype)
public:
	void clear() { *(uint64*)this = 0; }
};
 
class Decl : public DeclCommon {
public:
	enum OutType {
		OT_NONE=0,	// it's as if it never existed...
		OT_FROMHDR=1,	// came from an include file
		OT_HDR=2,	// destined for an include file
		OT_CODE=3,	// destined for main code body
		OT_REPOS=4,	// belongs in a code repository
		OT_GEN=5,	// was auto-generated
		OT_IMPLICIT=6,	// was implicitly added (but could have been added manually)
	};

private:
	bool		_defined:1;	ACCESSORS(Decl, bool, defined)
	bool		_initialized:1;	ACCESSOR(bool, initialized)
	bool		_condasn:1;	ACCESSORS(Decl, bool, condasn)
	bool		_anon:1;	ACCESSORS(Decl, bool, anon)
	bool		_newtype:1;	ACCESSORS(Decl, bool, newtype)
	bool		_field:1;	ACCESSORS(Decl, bool, field)
	bool		_bitfield:1;	ACCESSOR(bool, bitfield)
	OutType		_otype:3;	ACCESSORS(Decl, OutType, otype)

	// the following flags could be derived from the type:
	bool		_isfunc:1;	ACCESSORS(Decl, bool, isfunc)
	bool		_ismeth:1;	ACCESSORS(Decl, bool, ismeth)

	// flags for output dependency ordering:
	bool		_odefined:1;	ACCESSORS(Decl, bool, odefined)
	bool		_odeclared:1;	// also used to report duplicates
					ACCESSORS(Decl, bool, odeclared)
	bool		_oneeddef:1;	ACCESSORS(Decl, bool, oneeddef)
	bool		_oneeddecl:1;	ACCESSORS(Decl, bool, oneeddecl)

	type_t		_ptype;		// printable type (based on definition)
					ACCESSORS(Decl, type_t, ptype)
	type_t		_rtype;		// real type
					ACCESSORS(Decl, type_t, rtype)
	Scope		_scope;		ACCESSOR(Scope, scope)
	vStrTab::Id	_id;		ACCESSOR(vStrTab::Id, id)
	int32		_where;		ACCESSORS(Decl, int32, where)
	union {
		struct {
			ExprNode	*expr;
		} _v;			// for variable values
		struct {
			int64		val;
		} _c;			// for simple constant values
		struct {
			Decl		*decl;
		} _i;			// for inheritance
		struct {
			int		tsize;	// total size
			Decl		*first;
		} _a;			// for aggregates
		struct {
			int		bsize;	// bitsize
		} _b;			// for bitfields
		struct {
			vSymTab		*pscope;	// parameter scope
		} _f;			// for uninitialized functions
	};

	Decl 		*_next;	// next in-order declaration
	Decl 		*_link;	// link to other Decl's that share the same id 
	Decl		*_conv;	// what this has been converted to (if any)

	static csz* _otname[];

	void clear() { memset(this, 0, sizeof *this); }

public:
	Decl(const Decl &d) { *this = d; }
	Decl(const Decl &d, Scope sc) { *this = d; _scope = sc; }
	Decl(Scope sc, vStrTab::Id id) { clear(); _scope = sc; _id = id; }

	csz* cstr_id() const { return scope().strtab()->string(id()); }

	Expr vexpr() const { return Expr(_v.expr); }
	int64 cval() const { return _c.val; }
	Decl* idecl() const { return _i.decl; }
	int asize() const { return _a.tsize; }
	Decl* afirst() const { return _a.first; }
	int bbsize() const { return _b.bsize; }
	Decl& bitfieldsize(int n)
		{ _bitfield = true; _b.bsize = n; return *this; }
	Decl& idecl(Decl *d) { _i.decl = d; return *this; }
	Scope fpscope() const { return Scope(_f.pscope); }
	Decl& fpscope(Scope s) { _f.pscope = s; return *this; }

	static csz* otname(OutType o) { return _otname[o]; }
	csz* otname() const { return otname(otype()); }

	// TDecl& operator=(const TDecl &t)
	//	{ return static_cast<TDecl>(*this) = t; }

	Decl& vexpr(Expr e)
		{ _v.expr = e;
		  _defined = true; _initialized = true; return *this; }
	Decl& vexpr(StmtList s)
		{ _v.expr = Expr(s);
		  _defined = true; _initialized = true; return *this; }
	Decl& vexpr(ExprList e)
		{ _v.expr = Expr(e);
		  _defined = true; _initialized = true; return *this; }
	Decl& cval(int64 v)
		{ _c.val = v; _defined = true;
			_initialized = true; return *this; }
	bool purevirtual() const { return vexpr().op() == T(LITERAL); }
	bool isaggr() const { return rtype() && rtype()->istype(Type_Aggr); }
	bool nooutput(int mask)	const { return !((1<<otype()) & mask); }
	Decl& link(Decl *d) { _link = d; return *this; }
	Decl* link() const { return _link; }
	Decl* next() const { return _next; }
	Decl* nextfn() const { return !_link || _link->tagged() ? 0 : _link; }
	Decl* conv() const { return _conv; }
	Decl& conv(Decl *d) { _conv = d; return *this; }
	bool isctor() const // XXX - utilizing a major hack:
		{ return ismeth() && scope().strtab()->string(id())[0] == ' '; }
	bool implements(Decl *dt = 0) const;

	void add();
	void rename(vStrTab::Id id);
	void chain();
	void unchain();
	Decl *parent() const;
	Decl *clone(Scope, vStrTab::Id id = 0) const;

	bool istype() const;
	bool declarable() const;
};

// Note: constant values can be represented with Decl's with
// the appropriate type, and a default value (which will never change)


extern obj_t Type_DeclRef;

class DeclRef {
	Decl* _decl;
public:
	DeclRef(Decl *decl) : _decl(decl) {}
	type_t rtype() const { return _decl->rtype(); }
	Decl*  decl() const { return _decl; }
	static type_t type() { return &Type_DeclRef; }
};

class obj_DeclRef : public Object {
public:
	obj_DeclRef(Decl *decl) : Object(DeclRef::type(), Inline)
		{ new(&_val)DeclRef(decl); }
	GEN_COPYREFCTOR(obj_DeclRef, DeclRef)
};

GEN_DCAST(DeclRef)


extern obj_t Type_Enum;

class Enum {
	Decl *_decl;
public:
	Enum(Decl *decl) : _decl(decl) {}
	vStrTab::Id name() const { return _decl->id(); }
	Decl* decl() const { return _decl; }
	static type_t type() { return &Type_Enum; }
};

class obj_Enum : public Object {
public:
	obj_Enum(Decl *decl) : Object(Enum::type(), Inline)
		{ new(&_val)Enum(decl); }
	GEN_COPYREFCTOR(obj_Enum, Enum)
};

GEN_DCAST(Enum)


#define case_ANY_SCOPE_ID  \
			case AGGR_ID: case ENUM_ID: \
			case NSPACE_ID: case ASPACE_ID

#define case_ANY_TYPE_ID  case TYPE_ID: case AGGR_ID: \
			case ENUM_ID: case TPLATE_ID

#define case_ANY_VAL_ID	case VAL_ID: case CTOR_ID: \
			case DTOR_ID: case CONV_ID: case OPER_ID

#define case_ANY_DECL_ID case_ANY_TYPE_ID: case_ANY_VAL_ID: \
			 case NSPACE_ID: case ASPACE_ID

#define case_ANY_ID	case ID: case_ANY_DECL_ID



/* if this gives an error, then the size isn't 8 */
static const void * const test_sizeof_TDecl = sizeof(TDecl)-8;

Decl* symget(vStrTab::Id, const Scope, bool tagged);
Decl* symfind(vStrTab::Id, const Scope, bool tagged, bool search = true);

inline Decl*
Scope::uget(vStrTab::Id id) const
{ return symget(id, *this, false); }

inline Decl*
Scope::tget(vStrTab::Id id) const 
{ return symget(id, *this, true); }

inline Decl*
Scope::tget(vStrTab::Id id, bool t) const
{ return symget(id, *this, t); }

inline Decl*
Scope::ufind(vStrTab::Id id) const
{ return symfind(id, *this, false); }

inline Decl*
Scope::tfind(vStrTab::Id id) const
{ return symfind(id, *this, true); }

inline Decl*
Scope::tfind(vStrTab::Id id, bool t) const
{ return symfind(id, *this, t); }

inline Decl*
Scope::tfind(vStrTab::Id id, bool t, bool s) const
{ return symfind(id, *this,t,s); }


/* Note: it is assumed that these particular ID's won't occur naturally... */
const vStrTab::Id ID_ANON 	= (vStrTab::Id)-1;
const vStrTab::Id ID_INHERIT 	= (vStrTab::Id)-2;
const vStrTab::Id ID_USING	= (vStrTab::Id)-3;
const vStrTab::Id ID_FRIEND	= (vStrTab::Id)-4;
const vStrTab::Id ID_CONV	= (vStrTab::Id)-5;	// for conversion operators
const vStrTab::Id ID_CTOR	= (vStrTab::Id)-6;	// for conversion constructors

inline bool idreal(vStrTab::Id id) { return (ulong)id < (ulong)ID_CTOR; }

int tokenprec(vCAction::Token t, int assoc);


bool scope_empty(const Scope);
bool scope_derived(const Scope);
void dump_scope(const Scope);
void dump_scopes(ScopeList);

Decl * param_decl(Func *func, int i);

bool subscope(Scope scope, Scope sub);


// unfortunate arbitrary limits due to fixed size buffers 
// scattered throughout the code...  one of many things that
// needs to be fixed at some point in the future:
#define MAXTYPELEN 256	// (sometimes including id) 
#define MAXIDLEN 128	// (including all scoping)
#define MAXLINELEN 4096	// (for any single logical C++ line)

#endif
