/*  $Id: ParsePath.cc,v 1.4 1999/06/20 17:10:36 shaggy Exp $
 *      relative path: lib/parse
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define IMPLEMENTATION_ParsePath
#include <ParsePath>



void ParsePath::parse(const char *path)
{
const char *basep;
const char *dirp;
char *dp = dir;
char *bp = base;
	if((absolute = (path[0] == '/')))
		++path;
	dirp = basep = path;
	while(*path) {
		if(*path == '/')
			basep = path+1;
		++path;
	}
	while(dirp != basep)
		*dp++ = *dirp++;
	while(dp != dir && *(dp-1) == '/')
		dp--;
	*dp = 0;
	while(*basep)
		*bp++ = *basep++;
	*bp = 0;
	end = bp;
	next_dir = dir;
}


bool ParsePath::get_next_dir(char *nd)
{
	if(!*next_dir)
		return false;
	while(*next_dir == '/')
		++next_dir;
	if(nd) {
		while(*next_dir && *next_dir != '/')
			*nd++ = *next_dir++;
		*nd = 0;
	} else
		while(*next_dir && *next_dir != '/')
			next_dir++;
	return true;
}


bool ParsePath::has_suffix(const char *suffix) const
{
const char *bp = end;
const char *sp = suffix;
	while(*sp) {
		if(bp-- == base)
			return false;
		++sp;
	}
	while(*suffix) {
		if(*suffix != *bp)
			return false;
		++suffix;
		++bp;
	}
	return true;
}


bool ParsePath::change_suffix
		(char *nb, const char *olds, const char *news) const
{
const char *sufstart;
const char *bp = end;
const char *sp = olds;
	while(*sp) {
		if(bp-- == base)
			return false;
		++sp;
	}
	sufstart = bp;
	while(*olds) {
		if(*olds != *bp)
			return false;
		++olds;
		++bp;
	}
	bp = base;
	while(bp != sufstart)
		*nb++ = *bp++;
	while(*news)
		*nb++ = *news++;
	*nb++ = 0;
	return true;
}


const char* ParsePath::get_suffix(const char *div) const
{
const char *bp = end;
	while(bp-- != base) {
		const char *dp = div;
		const char *mp = bp;
		while(*mp == *dp) {
			mp++;
			dp++;
			if(!*dp)
				return bp;
		}
	}
	return 0;
}


const char* ParsePath::get_ext(const char *div) const
{
const char *cp;
	if(!(cp = get_suffix(div)))
		return 0;
	while(*div) {
		cp++;
		div++;
	}
	return cp;
}


void ParsePath::get_dir_path(char *np) const
{
const char *dp;
	dp = dir;
	while(dp != next_dir) 
		*np++ = *dp++;
	*np = 0;
}

